;*******************************************************************************
;
;	Filename:		DoorBell-1.asm
;	Date:				12 Aug 2017
;	File Version:	1.0
;	Author:			Larry Cicchinelli
;	Description:
;		This program uses a PIC 16F18313 and its Numerically Controlled Oscillator
;			It implements doorbell which uses multiple tones when it rings.
;
;
; Editor setting: spaces per indent and tabe size = 3
;	tabs are NOT expanded
;
;		PIC12F1501 I/O:
;			pin 1: VDD
;			pin 2: RA5						doorbell Sw1 - switches are LOW true
;			pin 3: RA4/AN3					doorbell Sw2
;			pin 4: RA3/~MCLR				doorbell Sw3
;			pin 5: RA2/ANA2				Tone output
;			pin 6: RA1/ANA1/ICSPCLK		Output to keep unit powered
;			pin 7: RA0/ANA0/ISCPDAT		not used
;			pin 8: VSS
;
; The frequency values are based on the CPU clock = 4MHz
; The min frequency is also the frequency step size.
; The max frequency is 1023 times the min frequency.
; The frequency range is 7.6Hz to 7.7KHz
;
; Definitions used by the program which are associated with specific I/O pins:
;	DO = digital output,  DI = digital input
;	DO_TONE			- tone output
;	DO_POWER			- keep power on
;	DI_SW0			- switches  for selecting tone to play
;	DI_SW1
;	DI_SW2
;
;*******************************************************************************
;*******************************************************************************

		#include p16F18313.inc
		
DEBUG				equ	0
MAX_LOOPS		equ	3		; nbr of time to repeat tone sequence if switch held

; CONFIG1
; __config 0xDF8F
 __CONFIG _CONFIG1, _FEXTOSC_OFF & _RSTOSC_HFINT32 & _CLKOUTEN_OFF & _CSWEN_ON & _FCMEN_OFF
; CONFIG2
; __config 0xF732
 __CONFIG _CONFIG2, _MCLRE_OFF & _PWRTE_OFF & _WDTE_OFF & _LPBOREN_OFF & _BOREN_OFF & _BORV_LOW & _PPS1WAY_OFF & _STVREN_ON & _DEBUG_OFF
; CONFIG3
; __config 0x3
 __CONFIG _CONFIG3, _WRT_OFF & _LVP_OFF
; CONFIG4
; __config 0x3
 __CONFIG _CONFIG4, _CP_OFF & _CPD_OFF


#if DEBUG == 0
DI_SW2			equ	RA5
DI_SW1			equ	RA4
DI_SW0			equ	RA3
DO_TONE			equ	RA2
DO_TONE_PPS		equ	RA2PPS
DO_POWER			equ	RA1
#else
DI_SW2			equ	RA3	; not available when debugging
DI_SW1			equ	RA1
DI_SW0			equ	RA0
DO_TONE			equ	RA2
DO_TONE_PPS		equ	RA2PPS
DO_POWER			equ	RA5
#endif
		
AI_MASK			equ	0
DI_MASK			equ	(1<<DI_SW0) | (1<<DI_SW1) | (1<<DI_SW2)
WPUA_MASK		equ	DI_MASK			; all DIs use weak pullups

TRISA_VAL		equ	AI_MASK | DI_MASK
ANSEL_VAL		equ	AI_MASK
RIGHT_JUSTIFY	equ	0x80 | (b'010'<<ADCS0)

;*******************************************************************************
; RAM
;*******************************************************************************

Access_RAM		udata_shr	0x70		; 16 bytes starting at address 0x70
Loop_Count		res	1
Delay_1			res	1
Delay_2			res	1
Switches			res	1
PrevNoteAddr	res	1

;*******************************************************************************
; Interrupt Vectors
;*******************************************************************************

; IMPORTANT: if either of these ISRs is changed, the test value for
; TABLE_END - TABLE_START must be changed to reflect the actual address of
; TABLE_START

	org	0									; Reset/PowerUp
	goto	Main

	org	0x04								; hardware interrupt
ISR.Exit
	retfie

; define the standard note lengths

L1.8		equ	.25			; 1/8 note in millisecs/10 - this defines the tempo!
L1.8P		equ	3*L1.8/2		; 1/8 note PLUS = 1/8 +(1/8)/2 = (3/8)/2
L1.4		equ	2*L1.8
L1.4P		equ	3*L1.4/2
L1.2		equ	4*L1.8
L1.2P		equ	3*L1.2/2
L1.16		equ	(L1.8/2)
L1.16P	equ	3*L1.16/2

DOT		equ	L1.8		; used for Morse Code
DASH		equ	3*L1.8

TABLE_START
; The information for the note frequencies was obtained from:
;		https://en.wikipedia.org/wiki/Scientific_pitch_notation
;--F5--
;  E5
;--D5--
;  C5
;--B4--
;  A4
;--G4--
;  F4
;--E4--
; define notes on the treble clef
;		table values are frequency * 0.524
; Fout = (CPU frequency/Accumulator Length/2) * Increment Value
;				(the /2 is due to using the fixed duty cycle (50%) feature)
;	Increment Value = Fout/(CPU frequency/Accumulator Length/2)
;						 = Fout/(4MHz/2^20/2)
;						 = Fout/1.907 = Fout*.524
; note: the table has the HIGH byte first due to the way the NCO needs to be
;			programmed with the LOW byte last.
; IMPORTANT: table MUST be completely in the first 256 bytes of program memory
REST	dw		-1, -1						; Rest indication, 2nd value is a dummy
C4		dw		HIGH	.137, LOW .137		; 262
D4		dw		HIGH	.154, LOW .154		; 294
E4		dw		HIGH	.173, LOW .173		; 330
F4		dw		HIGH	.183, LOW .183		; 349
F4#
G4b	dw		HIGH	.194, LOW .194		; 370
G4		dw		HIGH	.205, LOW .205		; 392
G4#
A4b	dw		HIGH	.217, LOW .217		; 415
MC_NOTE					; Morse Code Note
A4		dw		HIGH	.230, LOW .230		; 440
A4#
B4b	dw		HIGH	.244, LOW .244		; 466
B4		dw		HIGH	.259, LOW .259		; 494
C5		dw		HIGH	.274, LOW .274		; 523
C5#
D5b	dw		HIGH	.290, LOW .290		; 554
D5		dw		HIGH	.308, LOW .308		; 587
D5#
E5b	dw		HIGH	.326,	LOW .326		; 622
E5		dw		HIGH	.345, LOW .345		; 659
F5		dw		HIGH	.366, LOW .366		; 698
G5		dw		HIGH	.411, LOW .411		; 784
A5		dw		HIGH	.461, LOW .461		; 880
TABLE_END

#if TABLE_END - TABLE_START > .250	; 250 accounts for the two ISRs
	error "table too long"
#endif

; these tables do NOT have to be in low memory

Sequence_0	; Amazing Grace - 1st line
		dw		D4, L1.4
		dw		G4, L1.2
		dw		B4, L1.8
		dw		G4, L1.8
		dw		B4, L1.2
		dw		A4, L1.4
		dw		G4, L1.2
		dw		E4, L1.4
		dw		D4, L1.2
		dw		0					; terminate the list

Sequence_1	; When the Roll is Called Up Yonder - last line of chorus
		dw		D5b, L1.8P
		dw		B4b, L1.16
		dw		C5, L1.8P
		dw		C5, L1.16
		dw		C5, L1.8P
		dw		A4, L1.16
		dw		B4b, L1.8P
		dw		B4b, L1.16
		dw		B4b, L1.8P
		dw		G4, L1.16
		dw		A4b, L1.4P
		dw		0

Sequence_2
		dw		E4, L1.8
		dw		F4, L1.8
		dw		G4, L1.8
		dw		A4, L1.8
		dw		B4, L1.8
		dw		C5, L1.8
		dw		D5, L1.8
		dw		E5, L1.8
		dw		F5, L1.8
		dw		E5, L1.8
		dw		D5, L1.8
		dw		C5, L1.8
		dw		B4, L1.8
		dw		A4, L1.8
		dw		G4, L1.8
		dw		F4, L1.8
		dw		E4, L1.8
		dw		0

; Use Morse code numbers as the sequence for the remaining until changed
; by the user.
;
Sequence_3						; ...--
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DASH
		dw		REST, DOT
		dw		MC_NOTE, DASH
		dw		REST, DOT
		dw		0

Sequence_4						; ....-
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DASH
		dw		REST, DOT
		dw		0

Sequence_5						; .....
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		0
	
Sequence_6						; _....
		dw		MC_NOTE, DASH
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		0

Sequence_7						; --...
		dw		MC_NOTE, DASH
		dw		REST, DOT
		dw		MC_NOTE, DASH
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		MC_NOTE, DOT
		dw		REST, DOT
		dw		0
	
Sequence_Table
		dw		LOW Sequence_0,	HIGH Sequence_0
		dw		LOW Sequence_1,	HIGH Sequence_1
		dw		LOW Sequence_2,	HIGH Sequence_2
		dw		LOW Sequence_3,	HIGH Sequence_3
		dw		LOW Sequence_4,	HIGH Sequence_4
		dw		LOW Sequence_5,	HIGH Sequence_5
		dw		LOW Sequence_6,	HIGH Sequence_6
		dw		LOW Sequence_7,	HIGH Sequence_7


;*******************************************************************************
; MAIN PROGRAM
;*******************************************************************************

;MAIN_PROG CODE                      ; let linker place main program

Main
	call		init							; initialize the system hardware
	BANKSEL	LATA
	bsf		LATA, DO_POWER				; turn on power latch
	movlw		.2								; initiate a 20ms delay
	call		Delay.10x
	movlw		MAX_LOOPS					; set up nbr of times to repeat
	movwf		Loop_Count					;		tone sequence
	clrf		PrevNoteAddr				; no previous note

; read and process switches
;	move into bits 3..1 in order to make two-byte offset into Sequence_Table
	BANKSEL	PORTA
	movfw		PORTA							; read the switches
#if DEBUG == 0
	rrf		WREG, w						; put switch value
	rrf		WREG, w						;		into bits 3..1
#else
	bsf		WREG, 2						; force bit 2 high since RA3 not available 
	lslf		WREG, w						; move bits 2..0 to 3..1
#endif
	xorlw		b'1110'						; invert bits 3..1 (weak pullups enabled)
	andlw		b'1110'						; keep only bits 3..1
	movwf		Switches						; save the switch states
	
Main.10
	BANKSEL	LATA
	bsf		LATA, DO_POWER				; turn on power latch

; determine sequence to play using switch inputs
	movlw		HIGH Sequence_Table		; get high byte of table address
	bsf		WREG, 7						;		and set it for program memory
	movwf		FSR1H							; save for indirect access
	movlw		LOW Sequence_Table		; get low byte
	movwf		FSR1L							; save
; FSR1 = address of Sequence_Table
; now adjust for switch inputs
	movfw		Switches						; get switches value
	addwf		FSR1L, f						; get address of table entry for switch
	clrf		WREG							; set up to
	addwfc	FSR1H, f						;	update high byte of address
; FSR1 = address within Sequence_Table of sequence to play 
	moviw		FSR1++						; get low byte of address of notes to play
	movwf		FSR0L							; save for indirect access
	moviw		FSR1--						; get high byte of address
	movwf		FSR0H							; save
	bsf		FSR0H, 7						; force program memmory access
; FSR0 = address of the sequence
	call		Play							; play the tune/sequence

	BANKSEL	LATA
	bcf		LATA, DO_POWER				; turn off power latch

; the following code will only execute if a doorbell sw is still closed
	movlw		.100							; set up to
	call		Delay.10x					;		delay 1000ms (1s)

	decfsz	Loop_Count, f				; play sequence up to MAX_LOOPS times
	goto		Main.10

; This executes ONLY if a switch is pressed for longer than MAX_LOOPS
; iterations of the selected tone sequence

; set up for low power clock
	BANKSEL	OSCCON1
	movlw		(b'100'<<NOSC0) | (b'1001'<<NDIV0)
	movwf		OSCCON1						; LFINTOSC (31KHz), DIV = 512 = 60Hz
Main.100
	goto		Main.100
	
Play
; play a note sequence
; enter with:	FSR0 = address of start of tone sequence - must be in low memory
;					PrevNoteAddr = address of previously played note
; exit with:	FSR0 = address beyond end of tone sequence
; uses:			FSR0, FSR1
;
	movfw		INDF0							; get address of frequency value
	btfsc		STATUS, Z					; skip next if not done
	return
	subwf		PrevNoteAddr, w			; check if same note as last time
	btfss		STATUS, Z					; skip next if same note
	goto		Play.10
; handle same note as last one
	movlw		L1.8/4						; set up
	call		Delay.10x					;			 for 1/32 note delay
Play.10
	moviw		FSR0++						; get address of frequency value
	movwf		PrevNoteAddr				; save for next time
	call		Set.NCO
	moviw		FSR0++						; get the note/REST length
	call		Delay.10x
	call		NCO.Off
	goto		Play							; do next note	


Set.NCO
; set the NCO frequency
; enter with:	W = address of HIGH byte of frequency value in frequency table
; uses:			FSR1
; if frequency is negative, the "note" is a REST and the function will return
;		without programming the NCO
	movwf		FSR1L							; set address of table entry
	movlw		0x80							; set up to access
	movwf		FSR1H							;		program memory - in first 256 words!!
	moviw		FSR1++						; get high byte of table entry
	btfsc		WREG, 7						; skip next if bit 7=0: value is positive
	return									; must be a REST
; set the NCO frequency and enable it
	BANKSEL	NCO1INCU
	clrf		NCO1INCU
	movwf		NCO1INCH
	movfw		INDF1							; get low byte
	movwf		NCO1INCL						; this writes all 3 bytes to the register
	bsf		NCO1CON, N1EN				; enable the NCO
	return
; Set.NCO

	
NCO.Off
; synchronously disable the NCO output
; wait for output low
	BANKSEL	NCO1INCU
NCO.Off.10
	btfsc		NCO1CON, N1OUT				; br if output low
	goto		NCO.Off.10
; disable NCO output
	bcf		NCO1CON, N1EN				; disable the NCO
	return
; NCO.Off


Delay.10x
; Delay 10 times the nbr of milliseconds specified in the W register
;	The delay time is dependant on the CPU clock being 4MHz = 1MHz instruction
;	rate = 1us/inst.
; enter with: W = nbr of milliseconds to delay / 10
; uses: Delay_1, (Delay_2, W)

	movwf		Delay_1						; save the value
Delay.10x.10
	call		Delay.10ms
	decfsz	Delay_1, f
	goto		Delay.10x.10
	return
; Delay.10x

; this routine does NOT take into account the time to execute the individual
;	loop instructions or the calling sequence.  Therefore it will be somewhat
;	longer than 10ms.
; uses: W, Delay_2
Delay.10ms									; delay 10ms
	movlw		.10
	movwf		Delay_2						;  save loop iterations
Delay.10ms.10
	call		Delay.1ms					; 1ms
	decfsz	Delay_2, f					; 1 update counter
	goto		Delay.10ms.10				; 2
	return
; Delay.10ms	

; This routine is a delay of almost 1msec.
; 1ms delay = 1000 I-cycles with a 4MHz clock
; uses: W
; the typical calling sequence is:
;		call		Delay.1ms
; this instruction, plus the return, takes 4 I-cycles
;
; 4 + 1 + (123*8) + 7 = 984 + 12 = 996 I-cycles
Delay.1ms									; delay 1 msec
	call		Delay.1ms.100				; 4 I-cycles + calling and returning
												;		simulate 1 loop iteration
	movlw		.124							; 1 nbr of loops of 8 I-cycles each
; each iteration of the loop is 8 I-cycles except the last which is 7
Delay.1ms.50
	nop										; 1
	call		Delay.1ms.100				; 4 I-cycles
	decfsz	WREG, f						; 1 update counter
	goto		Delay.1ms.50				; 2
Delay.1ms.100
	return
; Delay.1ms


;===============================================================================
;									Initialize Hardware
;===============================================================================

init	; initialize the system hardware

; Oscillator - set to 4MHz
	BANKSEL	OSCCON1
	movlw		(b'110'<<NOSC0) | (b'0000'<<NDIV0)
	movwf		OSCCON1						; HFINTOSC = 1MHz, DIV = 1
												; OSCCON2 is read-only
	clrf		OSCCON3
	movlw		1<<HFOEN						; enable the HF internal oscillator
	movwf		OSCEN
	movlw		3								; select 4MHz
	movwf		OSCFRQ
init.10
	btfss		OSCCON3, NOSCR				; wait for oscillator ready
	goto		init.10

; Port A
	BANKSEL	PORTA
	clrf		PORTA
	BANKSEL	TRISA
	movlw		TRISA_VAL
	movwf		TRISA							; set analog and digital input pins
	BANKSEL	ANSELA
	movlw		TRISA							; set analog input pins
	movwf		ANSELA
	BANKSEL	INLVLA
	movlw		0xFF							; set up to
	movwf		INLVLA						;		make all inputs Schmitt Trigger CMOS
	BANKSEL	WPUA
	movlw		WPUA_MASK					; get the weak pullup mask
	movwf		WPUA

; Peripheral Pin Select
	BANKSEL	DO_TONE_PPS					; 0xE90
	movlw		b'11101'						; NCO output
	movwf		DO_TONE_PPS

; A/D convertor
	BANKSEL	ADCON0
	movlw		(1<<ADON)					; A/D on, used later to select A/D channel
	movwf		ADCON0
	movlw		RIGHT_JUSTIFY				; clk=Fosc/32, VRPOS=VDD, right justify data
	movwf		ADCON1
	clrf		ADACT							; no auto conversion trigger

; Interrupts
	clrf		INTCON						; clear global interrupt enable
	BANKSEL	PIE0
	clrf		PIE0							; clear peripheral interrupt enables
	clrf		PIE1							; clear peripheral interrupt enables

; NCO
	BANKSEL	NCO1ACCL
	clrf		NCO1CON						; set to Fixed Duty Cycle (50%)
	clrf		NCO1CLK						; use HFINTOSC for clock

	return
; init

	END
